/*
 * Copyright 2013 Yuichiro Moriguchi
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.morilib.sh.file;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.morilib.sh.AbstractShFileSystem;
import net.morilib.sh.AbstractShVirtualDirectory;
import net.morilib.sh.ShEnvironment;
import net.morilib.sh.ShFile;
import net.morilib.sh.ShSecurityPolicy;

public class ShWindowsFileSystem extends AbstractShFileSystem {

	private static final String SEP =
		System.getProperty("file.separator");
	private static final Pattern WPN =
		Pattern.compile("([A-Za-z]):[/\\\\]+(.*)");

	private final ShFile
	root = new AbstractShVirtualDirectory(null, null) {

		public Collection<ShFile> getFiles() {
			List<ShFile> l = new ArrayList<ShFile>();
			SortedSet<String> z;
			ShFile f;

			z = new TreeSet<String>();
			for(File g : File.listRoots()) {
				z.add(g.toString().substring(0, 1));
			}

			for(String s : z) {
				f = new ShWindowsRealFile(ShWindowsFileSystem.this,
						s.toLowerCase(),
						new File(s + ":"));
				l.add(f);
			}
			return l;
		}

	};

	private String home;
	private ShSecurityPolicy security;

	/**
	 * 
	 * @param home
	 * @param security
	 */
	public ShWindowsFileSystem(String home,
			ShSecurityPolicy security) {
		this.home = home;
		this.security = security;
	}

	/**
	 * 
	 * @param wpath
	 * @return
	 */
	public static String toJshFilename(String wpath) {
		StringBuffer b = new StringBuffer("/");
		Matcher m = WPN.matcher(wpath);
		String[] a;

		if(!m.matches())  return "";
		b.append(m.group(1).toLowerCase()).append('/');
		a = m.group(2).split("[/\\\\]+");
		for(String s : a)  b.append('/').append(s);
		return b.toString();
	}

	public ShFile getFile(String s) {
		StringBuffer b = new StringBuffer();
		String[] a;
		String t, v;

		if(s == null)  throw new NullPointerException();
		if(s.length() == 0)  return getCurrentDirectory();
		if(s.matches("/+")) {
			return root;
		} else if(s.charAt(0) == '/') {
			t = s.replaceFirst("^/+", "");
			t = t.replaceFirst("/+$", "");
			t = t.replaceAll("^(/\\.\\.)+", "/");
			do {
				v = t;
				t = t.replaceAll("/[^/]+/\\.\\.", "");
			} while(!t.equals(v));
			t = t.replaceAll("(/\\.)+/", "/");
			t = t.replaceAll("(/\\.)+$", "");
			a = t.split("/+");
			v = s.replaceAll("/+", "/");
			if(a[0].length() != 1 ||
					a[0].charAt(0) < 'a' || a[0].charAt(0) > 'z') {
				return null;
			} else {
				b.append('/').append(a[0].toUpperCase()).append(':');
				for(int i = 1; i < a.length; i++) {
					b.append(SEP).append(a[i]);
				}
				return new ShWindowsRealFile(
						this, v, new File(b.toString()));
			}
		} else if(realcur == null) {
			return getFile("/", s);
		} else {
			return getFile(getCurrentDirectory(), s);
		}
	}

	public ShFile getFile(String dir, String name) {
		return getFile(dir + "/" + name);
	}

	public ShFile getFile(ShFile dir, String name) {
		return getFile(dir + "/" + name);
	}

	public ShFile getRoot() {
		return root;
	}

	public ShSecurityPolicy getSecurityPolicy() {
		return security;
	}

	public InputStream getProfile() {
		ShFile f;

		try {
			if((f = getFile(home, "jsh_profile.sh")).isFile() ||
					(f = getFile(home, "jsh_login.sh")).isFile() ||
					(f = getFile(home, "profile.sh")).isFile()) {
				return f.getInputStream();
			} else {
				return null;
			}
		} catch (IOException e) {
			return null;
		}
	}

	public InputStream getRc() {
		ShFile f;

		try {
			if((f = getFile(home, "jshrc.sh")).isFile()) {
				return f.getInputStream();
			} else {
				return null;
			}
		} catch (IOException e) {
			return null;
		}
	}

	public ShFile getHome() {
		return getFile(home);
	}

	/* (non-Javadoc)
	 * @see net.morilib.sh.ShFileSystem#getNativeFile(java.lang.String)
	 */
	public ShFile getNativeFile(String s) {
		File x = new File(s);

		try {
			x = x.getCanonicalFile();
		} catch (IOException e) {
			// do nothing
		}
		return new ShWindowsRealFile(this, toJshFilename(x.toString()),
				x);
	}

	public ShFile searchPath(ShEnvironment env, String name) {
		List<String> p = env.getPath();
		ShFile f;

		if(name.length() == 0 || name.charAt(0) == '/') {
			return getFile(name);
		} else {
			for(String t : p) {
				if((f = getFile(t)) == null || !f.isDirectory()) {
					// do nothing
				} else if((f = getFile(t, name + ".exe")) == null &&
						(f = getFile(t, name + ".EXE")) == null) {
					// do nothing
				} else if(f.isExecutable()) {
					return f;
				}
			}
			return null;
		}
	}

}
